﻿//////////////////////////////////////////////
// BufferView.h
//
//////////////////////////////////////////////

/// Defines / Macros -------------------------

#pragma once

/// Includes ---------------------------------

// nkMemory
#include "../Dll/DllDefines.h"

#include "Buffer.h"
#include "BufferCast.h"

// Standards
#include <vector>

/// Class ------------------------------------

namespace nkMemory
{
	template <typename T = unsigned char>
	class BufferView
	{
		private :

			// Attributes
			T* _data ;
			unsigned long long _size ;

		public :

			// Constructor
			BufferView ()
			:	_data (nullptr),
				_size (0)
			{
				// Nothing to do
			}

			BufferView (T* data, unsigned long long size)
			:	_data (data),
				_size (size)
			{
				// Nothing to do
			}

			BufferView (const Buffer& buffer)
			:	_data ((T*)buffer.getData()),
				_size (buffer.getSize() / sizeof(T))
			{
				// Nothing to do
			}

			BufferView (const Buffer& buffer, unsigned long long index, unsigned long long length)
			:	_data ((T*)buffer.getData() + sizeof(T) * index),
				_size ((buffer.getSize() - index) / sizeof(T))
			{
				// Nothing to do
			}

			// Destructor
			// Doesn't free the data, merely a view
			~BufferView () = default ;

		public :

			// Getters
			T* getData () const
			{
				return _data ;
			}

			unsigned long long getSize () const
			{
				return _size ;
			}

			bool empty () const
			{
				return !_size ;
			}

			T& front ()
			{
				return _data[0] ;
			}

			T& back ()
			{
				return _data[_size - 1] ;
			}

		public :

			// Conversions
			BufferView<T> subView (unsigned long long index, unsigned long long length = 0) const
			{
				// Clamp size and length
				index = index < _size ? index : _size ;
				length = ((length > _size - index) || !length) ? _size - index : length ;

				if (!length)
					return BufferView<T>() ;

				return BufferView<T>(&_data[index], length) ;
			}

		public :

			// Operators
			T& operator[] (unsigned long long index)
			{
				return _data[index] ;
			}

			const T& operator[] (unsigned long long index) const
			{
				return _data[index] ;
			}

		public :

			// Constructors templated
			template <typename U>
			BufferView (BufferCast<U>& bufferCast)
			:	_data ((T*)bufferCast.getData()),
				_size ((bufferCast.getSize() * sizeof(U)) / sizeof(T))
			{
				// Check sizes to round to lower bounds
			}

			BufferView (BufferCast<T>& bufferCast)
			:	_data (bufferCast.getData()),
				_size (bufferCast.getSize())
			{
				// Check sizes to round to lower bounds
			}

			BufferView (std::vector<T>& vec)
			:	_data (vec.data()),
				_size (vec.size())
			{
				// Nothing to do
			}
	} ;
}

/// Exports ----------------------------------

namespace nkMemory
{
	// Unsigned
	// Unsigned
	template class DLL_MEMORY_EXPORT BufferView<unsigned char> ;
	template class DLL_MEMORY_EXPORT BufferView<unsigned short> ;
	template class DLL_MEMORY_EXPORT BufferView<unsigned int> ;
	template class DLL_MEMORY_EXPORT BufferView<unsigned long long> ;

	// Signed
	template class DLL_MEMORY_EXPORT BufferView<char> ;
	template class DLL_MEMORY_EXPORT BufferView<short> ;
	template class DLL_MEMORY_EXPORT BufferView<int> ;
	template class DLL_MEMORY_EXPORT BufferView<long long> ;

	// Floating
	template class DLL_MEMORY_EXPORT BufferView<float> ;
	template class DLL_MEMORY_EXPORT BufferView<double> ;
}